package cn.com.qmth.examcloud.web.interceptor;

import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.MDC;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.http.HttpStatus;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.method.HandlerMethod;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.ModelAndView;

import com.google.common.collect.Maps;
import com.google.common.collect.Sets;

import cn.com.qmth.examcloud.commons.helpers.FileChangeWatchdog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLogFactory;
import cn.com.qmth.examcloud.commons.logging.SLF4JImpl;
import cn.com.qmth.examcloud.commons.util.JsonUtil;
import cn.com.qmth.examcloud.commons.util.PathUtil;
import cn.com.qmth.examcloud.commons.util.StringUtil;
import cn.com.qmth.examcloud.commons.util.ThreadLocalUtil;
import cn.com.qmth.examcloud.web.support.ServletUtil;
import cn.com.qmth.examcloud.web.support.StatusResponse;

/**
 * 首发拦截器
 *
 * @author WANGWEI
 * @date 2018年5月22日
 * @Copyright (c) 2018-? http://qmth.com.cn All Rights Reserved.
 */
public class FirstInterceptor implements HandlerInterceptor {

	private static final ExamCloudLog LOG = ExamCloudLogFactory.getLog(FirstInterceptor.class);

	/**
	 * 接口日志
	 */
	protected static final ExamCloudLog INTERFACE_LOG = ExamCloudLogFactory
			.getLog("INTERFACE_LOGGER");

	private static final Set<String> RESOURCE_NAME_SET = Sets.newConcurrentHashSet();

	/**
	 * 拦截的请求
	 */
	private Set<String> badRequests;

	/**
	 * 热加载配置文件
	 *
	 * @author WANGWEI
	 * @param resourceName
	 */
	public synchronized void configureAndWatch(String resourceName) {
		if (RESOURCE_NAME_SET.contains(resourceName)) {
			return;
		}
		String path = PathUtil.getResoucePath(resourceName);
		loadFromFile(path);

		InterceptConfFileChangeWatchdog dog = new InterceptConfFileChangeWatchdog(path);
		dog.setDelay(30000);
		dog.start();
		RESOURCE_NAME_SET.add(resourceName);
	}

	/**
	 * 观察线程
	 *
	 * @author WANGWEI
	 */
	private class InterceptConfFileChangeWatchdog extends FileChangeWatchdog {

		protected InterceptConfFileChangeWatchdog(String path) {
			super(path);
		}

		@Override
		protected void doOnChange() {
			try {
				loadFromFile(path);
			} catch (Exception e) {
				LOG.info("Fail to load from file [" + path + "].", e);
			}
		}
	}

	private void loadFromFile(String path) {
		try {
			Set<String> newBadRequests = Sets.newHashSet();
			List<String> lines = FileUtils.readLines(new File(path), "UTF-8");
			for (String cur : lines) {
				if (StringUtils.isNotBlank(cur)) {
					newBadRequests.add(cur.trim());
				}
			}
			badRequests = newBadRequests;
			LOG.info("exclusions=" + JsonUtil.toJson(badRequests));
		} catch (IOException e) {
			LOG.error("fail to load config from file. filePath=" + path, e);
		}
	}

	@Override
	public boolean preHandle(HttpServletRequest request, HttpServletResponse response,
			Object handler) throws Exception {

		String traceID = request.getHeader("TRACE_ID");
		if (StringUtils.isBlank(traceID)) {
			traceID = ThreadLocalUtil.next();
		} else {
			ThreadLocalUtil.setTraceID(traceID);
		}

		// 设置MDC
		if (LOG instanceof SLF4JImpl) {
			MDC.put("TRACE_ID", traceID);
			MDC.put("KEY", "$$$$$");
		}

		String path = request.getServletPath();
		String method = request.getMethod();

		String reqPath = method + ":" + path;

		LOG.debug("[preHandle]. reqPath = " + reqPath);

		if (CollectionUtils.isNotEmpty(badRequests)) {
			if (badRequests.contains(reqPath)) {
				response.setStatus(HttpStatus.NOT_ACCEPTABLE.value());
				ServletUtil.returnJsonWithoutLog(new StatusResponse("406", "Not Acceptable."),
						response);
				return false;
			}
		}

		if (INTERFACE_LOG.isDebugEnabled()) {
			Map<String, String> headers = Maps.newHashMap();
			Enumeration<String> e = request.getHeaderNames();
			while (e.hasMoreElements()) {
				String name = (String) e.nextElement();
				String value = request.getHeader(name);
				if (name.startsWith("cookie")) {
					continue;
				}
				headers.put(name, value);
			}
			INTERFACE_LOG.debug(StringUtil.join("[preHandle]. path=\"", path, "\"; method=[",
					method, "]", "; headers=", JsonUtil.toJson(headers)));
		}

		if (handler instanceof HandlerMethod) {
			HandlerMethod handlerMethod = (HandlerMethod) handler;
			Object controllerObject = handlerMethod.getBean();
			RequestMapping mappingOfClass = AnnotationUtils
					.findAnnotation(controllerObject.getClass(), RequestMapping.class);
			RequestMapping mappingOfMethod = handlerMethod
					.getMethodAnnotation(RequestMapping.class);

			String[] allPathOfClass = null;
			String[] allPathOfMethod = null;

			if (null != mappingOfClass) {
				allPathOfClass = mappingOfClass.path();
			}
			if (null != mappingOfMethod) {
				allPathOfMethod = mappingOfMethod.path();
			}

			String pathsOfClass = null;
			String pathsOfMethod = null;

			if (null != allPathOfClass) {
				pathsOfClass = StringUtils.join(allPathOfClass, ",");
			}
			if (null != allPathOfMethod) {
				pathsOfMethod = StringUtils.join(allPathOfMethod, ",");
			}

			String mappingPath = StringUtils.join("[", pathsOfClass, "][", pathsOfMethod, "][",
					method, "]");

			if (INTERFACE_LOG.isDebugEnabled()) {
				INTERFACE_LOG.debug("[preHandle]. mapping = " + mappingPath);
			}

			request.setAttribute("$mappingPath", mappingPath);
			request.setAttribute("$ctrClass", controllerObject.getClass());

		} else {
			INTERFACE_LOG.error("cry.....");
			return false;
		}

		return true;
	}

	@Override
	public void postHandle(HttpServletRequest request, HttpServletResponse response, Object handler,
			ModelAndView modelAndView) throws Exception {
		LOG.debug("postHandle... ...");
	}

	@Override
	public void afterCompletion(HttpServletRequest request, HttpServletResponse response,
			Object handler, Exception ex) throws Exception {
		LOG.debug("afterCompletion... ...");
		// 清理MDC
		if (LOG instanceof SLF4JImpl) {
			MDC.clear();
		}
	}

}
