package cn.com.qmth.examcloud.web.security;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.MDC;
import org.springframework.http.HttpStatus;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.ModelAndView;

import com.google.common.collect.Sets;

import cn.com.qmth.examcloud.commons.logging.ExamCloudLog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLogFactory;
import cn.com.qmth.examcloud.commons.util.JsonUtil;
import cn.com.qmth.examcloud.commons.util.PathUtil;
import cn.com.qmth.examcloud.commons.util.RegExpUtil;
import cn.com.qmth.examcloud.web.RedisKeys;
import cn.com.qmth.examcloud.web.cloud.api.CloudService;
import cn.com.qmth.examcloud.web.cloud.api.EnterpriseService;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import cn.com.qmth.examcloud.web.security.bean.User;
import cn.com.qmth.examcloud.web.support.ServletUtil;
import cn.com.qmth.examcloud.web.support.StatusResponse;

/**
 * 请求鉴权
 *
 * @author WANGWEI
 * @date 2018年5月22日
 * @Copyright (c) 2018-? http://qmth.com.cn All Rights Reserved.
 */
public abstract class RequestPermissionInterceptor implements HandlerInterceptor {

	private static final ExamCloudLog LOG = ExamCloudLogFactory
			.getLog(RequestPermissionInterceptor.class);

	/**
	 * 接口日志
	 */
	protected static final ExamCloudLog INTERFACE_LOG = ExamCloudLogFactory
			.getLog("INTERFACE_LOGGER");

	private static final Set<String> RESOURCE_NAME_SET = Sets.newConcurrentHashSet();

	/**
	 * redis client
	 */
	private RedisClient redisClient;

	/**
	 * 例外请求前缀的正则表达式集合(不进行权限校验)
	 */
	private Set<String> exclusions;

	/**
	 * 热加载配置文件
	 *
	 * @author WANGWEI
	 * @param resourceName
	 */
	public synchronized void configure(String resourceName) {
		if (RESOURCE_NAME_SET.contains(resourceName)) {
			return;
		}
		String path = PathUtil.getResoucePath(resourceName);
		loadFromFile(path);

		RESOURCE_NAME_SET.add(resourceName);
	}

	/**
	 * 判断是否有请求权限
	 *
	 * @author WANGWEI
	 * @param mappingPath
	 * @param user
	 * @return
	 */
	public abstract boolean hasPermission(String mappingPath, User user);

	/**
	 * 构造函数
	 *
	 * @param redisClient
	 * @param exclusions
	 */
	public RequestPermissionInterceptor(RedisClient redisClient) {
		super();
		this.redisClient = redisClient;
	}

	@Override
	public boolean preHandle(HttpServletRequest request, HttpServletResponse response,
			Object handler) throws Exception {
		LOG.debug("preHandle... ...");

		Class<?> ctrClass = (Class<?>) request.getAttribute("$ctrClass");
		if (CloudService.class.isAssignableFrom(ctrClass)
				|| EnterpriseService.class.isAssignableFrom(ctrClass)) {
			return true;
		}

		String mappingPath = (String) request.getAttribute("$mappingPath");

		if ("[runtime][][GET]".equals(mappingPath)) {
			return true;
		}

		if (mappingPath.startsWith("[${server.error.path:${error.path:/error}}][]")) {
			response.setStatus(HttpStatus.NOT_FOUND.value());
			ServletUtil.returnJson(new StatusResponse("404", "not found."), response);
			return false;
		}

		if (CollectionUtils.isNotEmpty(exclusions)) {
			for (String e : exclusions) {
				if (mappingPath.matches(e)) {
					if (INTERFACE_LOG.isDebugEnabled()) {
						INTERFACE_LOG.debug("No authentication. mapping=" + mappingPath);
					}
					return true;
				}
			}
		}

		String key = null;
		String token = null;
		String kt = request.getHeader("user_token");
		if (StringUtils.isNotBlank(kt)) {
			String[] arr = kt.split(":");
			if (null != arr && 2 == arr.length) {
				key = arr[0];
				token = arr[1];
			}
		} else {
			key = request.getHeader("key");
			token = request.getHeader("token");
		}

		if (StringUtils.isBlank(key) && StringUtils.isBlank(token)) {
			key = request.getParameter("$key");
			token = request.getParameter("$token");
		}

		if (StringUtils.isBlank(key)) {
			response.setStatus(HttpStatus.FORBIDDEN.value());
			ServletUtil.returnJson(new StatusResponse("403", "key is blank."), response);
			return false;
		}
		if (StringUtils.isBlank(token)) {
			response.setStatus(HttpStatus.FORBIDDEN.value());
			ServletUtil.returnJson(new StatusResponse("403", "token is blank."), response);
			return false;
		}

		Integer sessionTimeout = redisClient.get(RedisKeys.SESSION_TIMEOUT, Integer.class);
		User user = null;
		if (null == sessionTimeout) {
			user = redisClient.get(key, User.class);
		} else {
			user = redisClient.get(key, User.class, sessionTimeout);
		}

		if (null == user) {
			response.setStatus(HttpStatus.FORBIDDEN.value());
			ServletUtil.returnJson(new StatusResponse("403", "no login."), response);
			return false;
		} else if (!token.equals(user.getToken())) {
			response.setStatus(HttpStatus.FORBIDDEN.value());
			ServletUtil.returnJson(new StatusResponse("403", "token is wrong."), response);
			return false;
		}

		MDC.put("KEY", key);

		if (!hasPermission(mappingPath, user)) {
			response.setStatus(HttpStatus.METHOD_NOT_ALLOWED.value());
			ServletUtil.returnJson(new StatusResponse("405", "no permission."), response);
			return false;
		}

		request.setAttribute("$accessUser", user);
		request.setAttribute("$kt", key + ":" + token);

		return true;
	}

	@Override
	public void postHandle(HttpServletRequest request, HttpServletResponse response, Object handler,
			ModelAndView modelAndView) throws Exception {
		LOG.debug("postHandle... ...");
	}

	@Override
	public void afterCompletion(HttpServletRequest request, HttpServletResponse response,
			Object handler, Exception ex) throws Exception {
		LOG.debug("afterCompletion... ...");
	}

	/**
	 * 加载配置项
	 *
	 * @author WANGWEI
	 * @param path
	 */
	private void loadFromFile(String path) {
		try {
			Set<String> newExclusions = Sets.newHashSet();
			List<String> lines = FileUtils.readLines(new File(path), "UTF-8");
			for (String cur : lines) {
				if (StringUtils.isNotBlank(cur)) {
					cur = cur.trim();
					if (cur.startsWith("regexp:")) {
						cur = StringUtils.replace(cur, "regexp:", "", 1);
					} else {
						cur = RegExpUtil.escape(cur);
					}
					newExclusions.add(cur.trim());
				}
			}
			exclusions = newExclusions;
			LOG.info("exclusions=" + JsonUtil.toJson(exclusions));
		} catch (IOException e) {
			LOG.error("fail to load config from file. filePath=" + path, e);
		}
	}

}
