package cn.com.qmth.examcloud.web.support;

import java.io.File;
import java.util.Map;
import java.util.Map.Entry;

import org.springframework.core.io.FileSystemResource;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestTemplate;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLogFactory;
import cn.com.qmth.examcloud.commons.util.JsonUtil;
import cn.com.qmth.examcloud.commons.util.StringUtil;
import cn.com.qmth.examcloud.commons.util.ThreadLocalUtil;
import cn.com.qmth.examcloud.web.cloud.api.BaseRequest;
import cn.com.qmth.examcloud.web.cloud.api.JsonSerializable;
import cn.com.qmth.examcloud.web.redis.RedisClient;

/**
 * 云服务客户端基类
 * 
 * @author WANGWEI
 *
 */
public abstract class CloudClientSupport {

	/**
	 * 接口日志
	 */
	protected static final ExamCloudLog INTERFACE_LOG = ExamCloudLogFactory
			.getLog("INTERFACE_LOGGER");

	protected abstract RestTemplate getRestTemplate();

	protected abstract RedisClient getRedisClient();

	protected abstract String getUrlPrefix();

	/**
	 * 获取响应体
	 * 
	 * @author WANGWEI
	 * @param respEntity
	 * @return
	 */
	private <T> T getRespBody(ResponseEntity<String> respEntity, Class<T> responseType) {

		String body = respEntity.getBody();
		if (HttpStatus.OK == respEntity.getStatusCode()) {
			if (null == responseType) {
				return null;
			}
			T t = JsonUtil.fromJson(body, responseType);
			return t;
		} else {
			StatusResponse sr = JsonUtil.fromJson(body, StatusResponse.class);
			throw new StatusException(sr.getCode(), sr.getDesc());
		}
	}

	/**
	 * 构建url
	 * 
	 * @param urlSuffix
	 * @return
	 */
	private String buildUrl(String urlSuffix) {
		String prefix = getUrlPrefix();
		prefix = prefix.endsWith("/") ? prefix : prefix + "/";
		urlSuffix = urlSuffix.startsWith("/") ? urlSuffix.substring(1) : urlSuffix;
		String url = prefix + urlSuffix;
		return url;
	}

	/**
	 * exchange
	 *
	 * @author WANGWEI
	 * @param url
	 * @param method
	 * @param body
	 * @param responseType
	 * @return
	 */
	protected <T> T exchange(String url, HttpMethod method, Object body, Class<T> responseType) {

		long startTime = System.currentTimeMillis();

		HttpHeaders requestHeaders = new HttpHeaders();
		requestHeaders.add("TRACE_ID", ThreadLocalUtil.getTraceID());
		requestHeaders.add("$spring_cloud_client", "_");
		getRedisClient().set("$_RMI_:" + ThreadLocalUtil.getTraceID(), System.currentTimeMillis(),
				10);

		if (INTERFACE_LOG.isInfoEnabled()) {
			INTERFACE_LOG.info("[CALL-IN]. url=" + url);
			if (body instanceof JsonSerializable) {
				INTERFACE_LOG.info("[CALL-REQ]. request=" + JsonUtil.toJson(body));
			}
		}

		HttpEntity<Object> requestEntity = new HttpEntity<Object>(body, requestHeaders);

		T respBody = null;
		try {
			ResponseEntity<String> respEntity = getRestTemplate().exchange(url, method,
					requestEntity, String.class);
			respBody = getRespBody(respEntity, responseType);

			String msg = StringUtil.join("[CALL-OK]. url=" + url,
					" ; cost " + (System.currentTimeMillis() - startTime)) + " ms.";
			INTERFACE_LOG.info(msg);
			INTERFACE_LOG.info("[CALL-RESP]. response=" + respEntity.getBody());
		} catch (StatusException e) {
			String msg = StringUtil.join("[CALL-FAIL]. url=" + url,
					" ; cost " + (System.currentTimeMillis() - startTime)) + " ms.";
			INTERFACE_LOG.error(msg);
			INTERFACE_LOG.error("[CALL-RESP]. response=" + e.toJson());
			throw e;
		} catch (Exception e) {
			String msg = StringUtil.join("[CALL-FATAL]. url=" + url, " ;", e.getMessage());
			INTERFACE_LOG.error(msg);
			throw e;
		}

		return respBody;
	}

	/**
	 * post请求
	 *
	 * @author WANGWEI
	 * @param urlSuffix
	 * @param body
	 * @param responseType
	 * @return
	 */
	protected <T> T post(String urlSuffix, BaseRequest body, Class<T> responseType) {
		String url = buildUrl(urlSuffix);
		return exchange(url, HttpMethod.POST, body, responseType);
	}

	/**
	 * post请求
	 * 
	 * @param urlSuffix
	 * @param body
	 */
	protected void post(String urlSuffix, BaseRequest body) {
		String url = buildUrl(urlSuffix);
		exchange(url, HttpMethod.POST, body, null);
	}

	/**
	 * 文件表单提交
	 *
	 * @author WANGWEI
	 * @param urlSuffix
	 * @param params
	 * @param file
	 * @param responseType
	 * @return
	 * @throws Exception
	 */
	public <T> T postForm(String urlSuffix, Map<String, String> params, File file,
			Class<T> responseType) {
		String url = buildUrl(urlSuffix);

		MultiValueMap<String, Object> param = new LinkedMultiValueMap<>();
		FileSystemResource resource = new FileSystemResource(file);
		param.add("file", resource);
		param.add("fileName", file.getName());

		for (Entry<String, String> entry : params.entrySet()) {
			param.add(entry.getKey(), entry.getValue());
		}

		return exchange(url, HttpMethod.POST, param, responseType);
	}

}
