package cn.com.qmth.examcloud.web.support;

import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Component;

import ch.qos.logback.classic.Logger;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLogFactory;
import cn.com.qmth.examcloud.commons.util.JsonUtil;
import cn.com.qmth.examcloud.commons.util.ObjectUtil;
import cn.com.qmth.examcloud.commons.util.StringUtil;
import cn.com.qmth.examcloud.commons.util.ThreadLocalUtil;
import cn.com.qmth.examcloud.web.cloud.api.BaseResponse;
import cn.com.qmth.examcloud.web.cloud.api.JsonSerializable;
import cn.com.qmth.examcloud.web.reports.BaseReport;
import cn.com.qmth.examcloud.web.reports.ReportLoggerFactory;

/**
 * spring mvc controller aspect.
 *
 * @author WANG WEI
 *
 */
@Component
@Aspect
public class ControllerAspect {

	/**
	 * 接口日志
	 */
	private static final ExamCloudLog INTERFACE_LOG = ExamCloudLogFactory
			.getLog("INTERFACE_LOGGER");

	public ControllerAspect() {
		System.out.println("=================ControllerAspect=================");
	}

	/**
	 * handlerMethods
	 *
	 * @author WANGWEI
	 */
	@Pointcut("@annotation(org.springframework.web.bind.annotation.RequestMapping) "
			+ "|| @annotation(org.springframework.web.bind.annotation.GetMapping) "
			+ "|| @annotation(org.springframework.web.bind.annotation.PostMapping)"
			+ "|| @annotation(org.springframework.web.bind.annotation.DeleteMapping)"
			+ "|| @annotation(org.springframework.web.bind.annotation.PutMapping)")
	public void handlerMethods() {

	}

	/**
	 * 接口处理
	 *
	 * @author WANG WEI
	 *
	 * @param joinPoint
	 * @return
	 */
	@Around("handlerMethods()")
	public Object doAroundWebRequests(ProceedingJoinPoint joinPoint) {
		long startTime = System.currentTimeMillis();

		// 获取request对象
		HttpServletRequest request = ServletUtil.getRequest();

		String path = request.getServletPath();
		String method = request.getMethod();

		// 请求信息
		String inMsg = StringUtil.join("[HTTP-IN]. path=\"", path, "\", method=[", method, "]");
		INTERFACE_LOG.info(inMsg);

		Object[] args = joinPoint.getArgs();

		if (INTERFACE_LOG.isInfoEnabled()) {
			if (null != args) {
				if (1 == args.length && args[0] instanceof JsonSerializable) {
					INTERFACE_LOG.info("[HTTP-REQ]. request=" + JsonUtil.toJson(args[0]));
				} else if (1 <= args.length) {
					StringBuilder sb = new StringBuilder();
					sb.append("[HTTP-REQ]. args: ");
					for (int i = 0; i < args.length; i++) {
						Object curArg = args[i];
						sb.append("args[").append(i).append("]=");
						if (null == curArg) {
							sb.append("null");
						} else if (curArg instanceof JsonSerializable) {
							sb.append(JsonUtil.toJson(curArg));
						} else if (ObjectUtil.isBaseDataType(curArg.getClass())) {
							sb.append(curArg);
						}
						sb.append("; ");
					}

					INTERFACE_LOG.info(sb.toString());
				}
			}
		}

		Object ret = null;
		boolean ok = false;
		try {

			// 执行
			ret = joinPoint.proceed();
			ok = true;
			long cost = System.currentTimeMillis() - startTime;

			String outMsg = StringUtil.join("[HTTP-OK]. path=\"", path, "\", method=[", method,
					"] ; cost ", cost, " ms.");

			INTERFACE_LOG.info(outMsg);

			if (null != ret && BaseResponse.class.isAssignableFrom(ret.getClass())) {
				BaseResponse baseResponse = (BaseResponse) ret;
				baseResponse.setCost(cost);
			}

		} catch (Throwable e) {
			String exceptionMsg = StringUtil.join("[HTTP-FAIL]. path=\"", path, "\", method=[",
					method, "] ; cost ", System.currentTimeMillis() - startTime, " ms.");
			INTERFACE_LOG.error(exceptionMsg);
			throw new RuntimeException(e);
		} finally {
			Boolean hasCollected = (Boolean) ThreadLocalUtil.get("$_HAS_COLLECTED");
			if (null != hasCollected && hasCollected) {
				String key = "$_REPORT_COLLECTIONS";
				@SuppressWarnings("unchecked")
				List<BaseReport> reportList = (List<BaseReport>) ThreadLocalUtil.get(key);
				for (BaseReport report : reportList) {
					Boolean ignoreException = report.getIgnoreException();
					if (null == ignoreException) {
						ignoreException = false;
					}
					if (ok || ignoreException) {
						Logger logger = ReportLoggerFactory.getLogger(report.getClass());
						report.setReportTime(new Date());
						report.setHasException(!ok);
						logger.info(JsonUtil.toJson(report) + "\n");
					}
				}
			}
		}

		if (INTERFACE_LOG.isInfoEnabled()) {
			boolean jsonSerializable = false;
			if (null == ret) {
				INTERFACE_LOG.info("[HTTP-RESP]. StatusCode=" + HttpStatus.OK);
				INTERFACE_LOG.info("[HTTP-RESP]. response=void");
			} else if (ret instanceof ResponseEntity) {
				ResponseEntity<?> re = (ResponseEntity<?>) ret;
				Object body = re.getBody();
				if (null != body) {
					if (body instanceof JsonSerializable) {
						jsonSerializable = true;
					} else if (body instanceof Collection) {
						jsonSerializable = true;
					} else if (body instanceof Map) {
						jsonSerializable = true;
					} else if (body instanceof String) {
						jsonSerializable = true;
					}
					INTERFACE_LOG.info("[HTTP-RESP]. StatusCode=" + re.getStatusCodeValue());
					if (jsonSerializable) {
						INTERFACE_LOG.info("[HTTP-RESP]. response=" + JsonUtil.toJson(body));
					}
				}
			} else {
				if (ret instanceof JsonSerializable) {
					jsonSerializable = true;
				} else if (ret instanceof Collection) {
					jsonSerializable = true;
				} else if (ret instanceof Map) {
					jsonSerializable = true;
				} else if (ret instanceof String) {
					jsonSerializable = true;
				}
				INTERFACE_LOG.info("[HTTP-RESP]. StatusCode=" + HttpStatus.OK);
				if (jsonSerializable) {
					INTERFACE_LOG.info("[HTTP-RESP]. response=" + JsonUtil.toJson(ret));
				}
			}
		}

		return ret;
	}

}